# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import os

import scipy.interpolate as si
from numpy import power, logical_and, inf, pi, exp, tanh, cos, sin, square, ones_like
from pandas import read_csv

from gwbench.utils import cLight

noise_curves_path=os.path.join(os.path.dirname(os.path.abspath(__file__)),'noise_curves')

def psd(tec,f,F_lo=-inf,F_hi=inf,psd_file=None,is_asd=None):

    # limit the PSD to a certain freq range
    f=f[logical_and(f>=F_lo,f<=F_hi)]

    # if a psd file is handed over, reset the tec label
    if psd_file is not None: tec = ''

    # use analytical PSDs, f_lo, f_hi will be that of freq array f
    if   tec == 'tec':
        f_lo = max(1.,F_lo)
        f_hi = min(16384.,F_hi)
        check_f(tec,f,f_lo,f_hi)
        return  1e-60 * ones_like(f[logical_and(f>=f_lo,f<=f_hi)]), f[logical_and(f>=f_lo,f<=f_hi)]
    elif tec == 'aLIGO':
        f_lo = max(10.,F_lo)
        f_hi = min(2048.,F_hi)
        check_f(tec,f,f_lo,f_hi)
        return  psd_aLIGO(f[logical_and(f>=f_lo,f<=f_hi)]), f[logical_and(f>=f_lo,f<=f_hi)]
    elif tec == 'CEwb':
        f_lo = max(5.,F_lo)
        f_hi = min(2048.,F_hi)
        check_f(tec,f,f_lo,f_hi)
        return  psd_CEwb(f[logical_and(f>=f_lo,f<=f_hi)]), f[logical_and(f>=f_lo,f<=f_hi)]
    elif tec == 'LISA-17':
        f_lo = F_lo
        f_hi = F_hi
        check_f(tec,f,f_lo,f_hi)
        return psd_LISA_17(f[logical_and(f>=f_lo,f<=f_hi)]), f[logical_and(f>=f_lo,f<=f_hi)]
    elif tec == 'LISA-Babak17':
        f_lo = F_lo
        f_hi = F_hi
        check_f(tec,f,f_lo,f_hi)
        return psd_LISA_Babak17(f[logical_and(f>=f_lo,f<=f_hi)]), f[logical_and(f>=f_lo,f<=f_hi)]
    elif 'LISA-Robson18' in tec:
        f_lo = F_lo
        f_hi = F_hi
        check_f(tec,f,f_lo,f_hi)
        curve = tec.split('-')[-1]
        if   curve == 'gen': curve = 0
        elif curve == '6mo': curve = 1
        elif curve == '1yr': curve = 2
        elif curve == '2yr': curve = 3
        elif curve == '4yr': curve = 4
        else: raise ValueError('Specified Robson18 curve not known, choose from: "gen", "6mo", "1yr", "2yr", "4yr".')
        return psd_LISA_Robson18(f[logical_and(f>=f_lo,f<=f_hi)],curve), f[logical_and(f>=f_lo,f<=f_hi)]

    # use PSDs from noise_curves files
    else:
        if tec == '':
            filename = psd_file
            asd = is_asd
        else:
            filename, asd = get_filename(tec)
            filename = os.path.join(noise_curves_path,filename)

        psd_file = read_csv(filename, sep = None, header = None, engine = 'python', comment = '#')
        psd_data = psd_file.to_numpy()
        psd = si.interp1d(psd_data[:,0], psd_data[:,1]**(1+asd))

        # find correct limits: file vs user-set limits
        f_lo = max(psd_data[0,0],F_lo)
        f_hi = min(psd_data[-1,0],F_hi)
        check_f(tec,f,f_lo,f_hi)

        # return the PSD and corresponding freq array
        return psd(f[logical_and(f>=f_lo,f<=f_hi)]), f[logical_and(f>=f_lo,f<=f_hi)]

def psd_aLIGO(f):
    x = f/245.4
    return 1.e-48 * ( 0.0152 * power(x,-4.) + 0.2935 * power(x,9./4) +
                2.7951 * power(x,3./2) - 6.5080 * power(x,3./4) + 17.7622 )

def psd_CEwb(f):
    return 5.623746655206207e-51 + 6.698419551167371e-50 * power(f,-0.125) + 7.805894950092525e-31 * power(f,-20.) + 4.35400984981997e-43 * power(f,-6.) \
            + 1.630362085130558e-53 * f + 2.445543127695837e-56 * square(f) + 5.456680257125753e-66 * power(f,5)

def psd_LISA_17(f):
    a1 = 8.2047564e-33
    a2 = 3.0292821e-38
    a3 = 1.4990886e-39
    a4 = 1.0216062e-40
    return a1 * (f/10**-4)**-6 + 0.8 * a2 * (f/(10**-3))**-4 + a3 * (f/0.1)**2 + a4

def psd_LISA_Babak17(f):
    L     = 2.5e9
    SnLoc = 2.89e-24
    SnSN  = 7.92e-23
    SnOmn = 4.00e-24
    SnAcc = (9.e-30 + 3.24e-28 * ((3.e-5/f)**10 + (1.e-4/f)**2)) * (2.*pi*f)**-4
    SGal  = 3.266e-44 * f**(-7./3.) * exp(-(f/1.426e-3)**1.183) * 0.5 * (1. + tanh(-(f-2.412e-3)/4.835e-3))
    return SGal + 20./3. * (4 * SnAcc + 2 * SnLoc + SnSN + SnOmn)/L**2 * (1 + (( 2*L*f / (0.41*cLight))**2 ))

def psd_LISA_Robson18(f,curve):
    L     = 2.5e9
    fstar = 19.09e-3
    Poms  = (1.5e-11)**2 * (1. + (2.e-3/f)**4)
    Pacc  = (3.e-15)**2  * (1. + (0.4e-3/f)**2) * (1. + (f/8.e-3)**4)

    if curve == 0:
        Snoc =  10./(3. * L**2) * (Poms + 4./(2. * pi * f)**4 * Pacc)                         * (1. + 0.6 * (f/fstar)**2)
        return  10./(3. * L**2) * (Poms + 2./(2. * pi * f)**4 * (1. + cos(f/fstar)**2) * Pacc) * (1. + 0.6 * (f/fstar)**2)
    else:
        if curve == 1:
            alpha = 0.133
            beta  = 243.
            kappa = 482.
            gamma = 917.
            fk    = 0.00258
        elif curve == 2:
            alpha = 0.171
            beta  = 292.
            kappa = 1020.
            gamma = 1680.
            fk    = 0.00215
        elif curve == 3:
            alpha = 0.165
            beta  = 299.
            kappa = 611.
            gamma = 1340.
            fk    = 0.00173
        elif curve == 4:
            alpha = 0.138
            beta  = -221.
            kappa = 521.
            gamma = 1680.
            fk    = 0.00113
        Snoc =  10./(3. * L**2) * (Poms + 4./(2. * pi * f)**4 * Pacc) * (1. + 0.6 * (f/fstar)**2)
        Sc   = 9.e-45 * f**(-7./3.) * exp(-f**alpha + beta * f * sin(kappa * f)) * (1. + tanh(gamma * (fk - f)))
        return Snoc + Sc

def get_filename(tec):
    # current Cosmic Explorer curves, see https://dcc.cosmicexplorer.org/CE-T2000017/public
    if   tec == 'CE-40':
        filename = 'cosmic_explorer_40km.txt'
        asd = 1
    elif tec == 'CE-40-LF':
        filename = 'cosmic_explorer_40km_lf.txt'
        asd = 1
    elif tec == 'CE-20':
        filename = 'cosmic_explorer_20km.txt'
        asd = 1
    elif tec == 'CE-20-PM':
        filename = 'cosmic_explorer_20km_pm.txt'
        asd = 1
    # https://apps.et-gw.eu/tds/?content=3&r=18213 --> 1st (frequencies) and 4th (xylophone PSD) columns of ET10kmcolumns.txt
    elif tec == 'ET-10-XYL':
        filename = 'et_10km_xylophone.txt'
        asd = 0
    # https://dcc.ligo.org/LIGO-T2300041-v1/public
    elif tec == 'A#':
        filename = 'a_sharp.txt'
        asd = 1
    # curves used in the trade study for the Cosmic Explorer Horizon Study, see https://dcc.cosmicexplorer.org/CE-T2000007/public
    elif tec == 'A+':
        filename = 'a_plus.txt'
        asd = 1
    elif tec == 'V+':
        filename = 'advirgo_plus.txt'
        asd = 1
    elif tec == 'K+':
        filename = 'kagra_plus.txt'
        asd = 1
    elif tec == 'Voyager-CBO':
        filename = 'voyager_cb.txt'
        asd = 1
    elif tec == 'Voyager-PMO':
        filename = 'voyager_pm.txt'
        asd = 1
    elif tec == 'ET':
        filename = 'et.txt'
        asd = 1
    elif tec == 'CE1-10-CBO':
        filename = 'ce1_10km_cb.txt'
        asd = 1
    elif tec == 'CE1-20-CBO':
        filename = 'ce1_20km_cb.txt'
        asd = 1
    elif tec == 'CE1-30-CBO':
        filename = 'ce1_30km_cb.txt'
        asd = 1
    elif tec == 'CE1-40-CBO':
        filename = 'ce1_40km_cb.txt'
        asd = 1
    elif tec == 'CE2-10-CBO':
        filename = 'ce2_10km_cb.txt'
        asd = 1
    elif tec == 'CE2-20-CBO':
        filename = 'ce2_20km_cb.txt'
        asd = 1
    elif tec == 'CE2-30-CBO':
        filename = 'ce2_30km_cb.txt'
        asd = 1
    elif tec == 'CE2-40-CBO':
        filename = 'ce2_40km_cb.txt'
        asd = 1
    elif tec == 'CE1-10-PMO':
        filename = 'ce1_10km_pm.txt'
        asd = 1
    elif tec == 'CE1-20-PMO':
        filename = 'ce1_20km_pm.txt'
        asd = 1
    elif tec == 'CE1-30-PMO':
        filename = 'ce1_30km_pm.txt'
        asd = 1
    elif tec == 'CE1-40-PMO':
        filename = 'ce1_40km_pm.txt'
        asd = 1
    elif tec == 'CE2-10-PMO':
        filename = 'ce2_10km_pm.txt'
        asd = 1
    elif tec == 'CE2-20-PMO':
        filename = 'ce2_20km_pm.txt'
        asd = 1
    elif tec == 'CE2-30-PMO':
        filename = 'ce2_30km_pm.txt'
        asd = 1
    elif tec == 'CE2-40-PMO':
        filename = 'ce2_40km_pm.txt'
        asd = 1
    else: raise ValueError(f'Specified PSD "{tec}" not known, choose from {tecs}.')

    return filename, asd

def check_f(tec,f,f_lo,f_hi):
    if f[-1] < f_lo: raise ValueError('The maximum frequency is below the low-frequency cutoff of the PSD for '+tec,f[-1],f_lo)
    if f[0]  > f_hi: raise ValueError('The minimum frequency is above the high-frequency cutoff of the PSD for '+tec,f[-1],f_hi)



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import logging
import os
import sys
from logging import getLevelName

import numpy as np

################################################################################
# available detectors technologies and locations
################################################################################

available_tecs = (
    'aLIGO', 'A+', 'V+', 'K+',
    'A#', 'Voyager-CBO', 'Voyager-PMO',
    'ET', 'ET-10-XYL', 'CEwb',
    'CE-40', 'CE-40-LF', 'CE-20', 'CE-20-PM',
    'CE1-10-CBO', 'CE1-20-CBO', 'CE1-30-CBO', 'CE1-40-CBO',
    'CE2-10-CBO', 'CE2-20-CBO', 'CE2-30-CBO', 'CE2-40-CBO',
    'CE1-10-PMO', 'CE1-20-PMO', 'CE1-30-PMO', 'CE1-40-PMO',
    'CE2-10-PMO', 'CE2-20-PMO', 'CE2-30-PMO', 'CE2-40-PMO',
    'LISA-17', 'LISA-Babak17', 'LISA-Robson18'
    )

available_locs = (
    'H', 'L', 'V', 'K', 'I', 'LHO', 'LLO', 'LIO',
    'ET1', 'ET2', 'ET3', 'ETS1', 'ETS2', 'ETS3',
    'C', 'N', 'S', 'CEA', 'CEB', 'CES'
    )

################################################################################
# constants
################################################################################

#-----constants in SI units-----
GNewton    = 6.6743e-11
cLight     = 2.99792458e8
Msun       = 1.9884099021470415e+30
Mpc        = 3.085677581491367e+22
REarth     = 6378136.6
AU         = 1.4959787066e11
year       = 3.1536e7
hPlanck    = 6.62607015e-34
TEarth     = 86400.
halfTEarth = TEarth / 2

#-----convert mass in solar masses to seconds-----
MTsun    = Msun * GNewton/cLight**3.
time_fac = MTsun
#-----convert mass/distance in solar mass/Mpc to dimensionless-----
strain_fac = GNewton/cLight**2.*Msun/Mpc


################################################################################
# basic functions
################################################################################

#-----combine two sympy symbols strings without duplicates-----
def reduce_symbols_strings(string1,string2):
    # combine the two lists
    symbols_list = string1.split(' ') + string2.split(' ')
    # remove duplicates
    symbols_list = list(dict.fromkeys(symbols_list))

    # recreate a string of symbols and return it
    return ' '.join(symbols_list)

#-----delete symbols in sympy symbols strings, if not present in the other-----
def remove_symbols(string1,string2,keep_same=1):
    symbs_list1 = string1.split(' ')
    symbs_list2 = string2.split(' ')
    # remove unwanted symbols from 1
    if keep_same:
        symbols_list = [x for x in symbs_list1 if x in symbs_list2]
    else:
        symbols_list = [x for x in symbs_list1 if x not in symbs_list2]

    # recreate a string of symbols and return it
    return ' '.join(symbols_list)

#-----get subarray-----
def get_sub_array_ids(arr,sub_arr):
    return np.logical_and(arr>=sub_arr[0],arr<=sub_arr[-1])

#-----use only subset of dictionary-----
def get_sub_dict(dic, key_list, keep_in_list=1):
    if type(key_list) == str: key_list = key_list.split(' ')
    if keep_in_list: return {k:v for k,v in dic.items() if k     in key_list}
    else:            return {k:v for k,v in dic.items() if k not in key_list}

#-----check if list is subset of another list-----
def is_subset_lists(sub, sup):
    return all([el in sup for el in sub])


################################################################################
# waveform manipluations
################################################################################

#-----get amp/pha for pl and cr polarizations (since they are complex)-----
def transform_hfpc_to_amp_pha(hfpc, f, params_list):
    hfp, hfc = hfpc(f, *params_list)
    return pl_cr_to_amp_pha(hfp, hfc)

def pl_cr_to_amp_pha(hfp, hfc):
    hfp_amp, hfp_pha = amp_pha_from_z(hfp)
    hfc_amp, hfc_pha = amp_pha_from_z(hfc)
    return hfp_amp, hfp_pha, hfc_amp, hfc_pha

#-----convert amp/phase derivatives to re/im ones-----
def z_deriv_from_amp_pha(amp,pha,del_amp,del_pha):
    del_z = np.zeros(del_amp.shape,dtype=np.complex_)
    if len(del_amp.shape) == 2:
        for i in range(del_amp.shape[1]):
            del_z[:,i] = del_amp[:,i] * np.exp(1j*pha) + amp * np.exp(1j*pha) * 1j * del_pha[:,i]
        return del_z
    else:
        return del_amp * np.exp(1j*pha) + amp * np.exp(1j*pha) * 1j * del_pha

#-----re/im vs. amp/phase transformations-----
def re_im_from_amp_pha(amp,pha):
    return re_im_from_z(z_from_amp_pha(amp,pha))

def amp_pha_from_re_im(re,im):
    return amp_pha_from_z(z_from_re_im(re,im))

#-----re/im or amp/phase vs. complex number transformations-----
def re_im_from_z(z):
    return np.real(z), np.imag(z)

def z_from_re_im(re,im):
    return re + 1j * im

def amp_pha_from_z(z):
    return np.abs(z), np.unwrap(np.angle(z))

def z_from_amp_pha(amp,pha):
    return amp * np.exp(1j*pha)


################################################################################
# IO functions
################################################################################

#-----Block and unblock printing-----
def block_print(active=1):
    if active: sys.stdout = open(os.devnull, 'w')
    return

def unblock_print(active=1):
    if active: sys.stdout = sys.__stdout__
    return

#-----sending warning or error message-----
def log_msg(message, logger=None, level='INFO'):
    if logger is None: print(level + ': ' + message)
    else:              logger.log(getLevelName(level), message)
    if level in ['ERROR', 'CRITICAL']: sys.exit()

def get_logger(name, level='INFO', stdout=True, logfile=None):
    logging.basicConfig(format = '%(asctime)s - %(name)s - %(levelname)s : %(message)s')
    if stdout: logging.basicConfig(stream = sys.stdout)
    if logfile is not None: logging.basicConfig(filename = logfile, filemode = 'w')
    logger = logging.getLogger(name)
    set_logger_level(logger, level)
    return logger

def set_logger_level(logger, level):
    logger.setLevel(level)


################################################################################
# network_spec handlers
################################################################################

def read_det_keys_from_label(network_label):
    det_keys = []

    ##-----read the network label and find all detectors-----
    keys = list(network_label)

    # - in network list means that all 2G detectors up to that index are to be
    # taken at aLIGO sensitivity
    aLIGO = int('-' in keys)
    if aLIGO: aLIGO_id = keys.index('-')

    # + in network list means that all 2G detectors up to that index are to be
    # taken at A+ sensitivity
    a_pl = int('+' in keys)
    if a_pl: a_pl_id = keys.index('+')

    # v in network list means that all 2G detectors up to that index are to be
    # taken at Voyager sensitivity
    voy = int('v' in keys)
    if voy:
        voy_id = keys.index('v')
        tmp = int(keys[voy_id+1] == 'p')
        voy_pmo = tmp * 'PMO' + (1-tmp) * 'CBO'

    # find out which locations with which PSDs are in the network
    for loc in available_locs:
        if loc in keys:
            loc_id = keys.index(loc)

            if loc in ('H','L','V','K','I'):
                if aLIGO and loc_id < aLIGO_id:
                    name = 'aLIGO_'+loc
                elif a_pl and loc_id < a_pl_id:
                    if loc == 'V':
                        name = 'V+_'+loc
                    elif loc == 'K':
                        name = 'K+_'+loc
                    else:
                        name = 'A+_'+loc
                elif voy and loc_id < voy_id:
                    name = 'Voyager-{}_{}'.format(voy_pmo,loc)

            elif loc in ('C','N','S'):
                if keys[loc_id+1] == 'c':
                    name = f'CE-{keys[loc_id+2]}0'
                    if   keys[loc_id+2] == 'l': name += '-LF'
                    elif keys[loc_id+2] == 'p': name += '-PM'
                    name += f'_{loc}'
                else:
                    ce_a = int(keys[loc_id+1] == 'a') # 0 for i, 1 for a - CE1 as i, CE2 as a
                    ce_arm = int(keys[loc_id+2])*10  # arm length (n for n*10km)
                    tmp = int(keys[loc_id+3] == 'p')
                    ce_pmo = tmp * 'PMO' + (1-tmp) * 'CBO'
                    name = f'CE{ce_a+1}-{ce_arm}-{ce_pmo}_{loc}'

            det_keys.append(name)

    # add 3 ET detectors
    if 'E' in keys:
        for name in ['ET_ET1','ET_ET2','ET_ET3']:
            det_keys.append(name)

    return det_keys



import numpy as np

from gwbench.antenna_pattern_np import ant_pat_vectors, calc_gra, det_ten_and_loc_vec, loc_phase_func
from gwbench.utils import REarth, halfTEarth

cos = np.cos
sin = np.sin

# for the antenna pattern and location phase derivatives
# del__ra_dec_psi_tc__gra/dec/psi
deriv_keys    = ('ra', 'dec', 'psi', 'tc')

def d_gra_dec_psi(use_rot):
    if use_rot is not None and use_rot:
        return np.array([ [ -1, 0, 0, np.pi / halfTEarth ],
                          [  0, 1, 0, 0 ],
                          [  0, 0, 1, 0 ] ])
    else:
        return np.array([ [ -1, 0, 0, 0 ],
                          [  0, 1, 0, 0 ],
                          [  0, 0, 1, 0 ] ])

def dr_ana_derivs(f, params_dic, hf, hfp, hfc, Fp, Fc, Flp, loc, use_rot, user_locs=None):
    d_hf              = wf_ana_derivs(f, params_dic, hf)
    d_Fp, d_Fc, d_Flp = antenna_pattern_and_loc_phase_fac_ana_derivs(f, params_dic, loc, use_rot, user_locs=user_locs)
    return {
        'DL'    : d_hf['DL'],
        'phic'  : d_hf['phic'],
        'tc'    : d_Flp['tc']  * (hfp * Fp + hfc * Fc) + Flp * (hfp * d_Fp['tc']  + hfc * d_Fc['tc'] ) + d_hf['tc'],
        'ra'    : d_Flp['ra']  * (hfp * Fp + hfc * Fc) + Flp * (hfp * d_Fp['ra']  + hfc * d_Fc['ra'] ),
        'dec'   : d_Flp['dec'] * (hfp * Fp + hfc * Fc) + Flp * (hfp * d_Fp['dec'] + hfc * d_Fc['dec']),
        'psi'   : d_Flp['psi'] * (hfp * Fp + hfc * Fc) + Flp * (hfp * d_Fp['psi'] + hfc * d_Fc['psi']),
        }

# waveform derivatives
def wf_ana_derivs(f, params_dic, hf):
    return {
        'DL'    : -hf / params_dic['DL'],
        'tc'    : (1j * 2 * np.pi) * f * hf,
        'phic'  : -1j * hf,
        }

# antenna pattern and location phase factor derivatives
def antenna_pattern_and_loc_phase_fac_ana_derivs(f, params_dic, loc, use_rot, user_locs=None):
    gra        = calc_gra(params_dic['ra'], params_dic['gmst0'], use_rot, f, params_dic['Mc'], params_dic['tc'])
    D, d       = det_ten_and_loc_vec(loc, REarth, user_locs=user_locs)
    d_XX, d_YY = np.matmul(calc_d_ant_pat_vectors(gra, params_dic['dec'], params_dic['psi']), d_gra_dec_psi(use_rot))
    d_Fp, d_Fc = calc_d_ant_pat_funcs(D, *ant_pat_vectors(gra, params_dic['dec'], params_dic['psi']))
    return { key : deriv for key,deriv in zip(deriv_keys, (np.matmul(d_Fp['XX'], d_XX) + np.matmul(d_Fp['YY'], d_YY)).flatten()) }, \
           { key : deriv for key,deriv in zip(deriv_keys, (np.matmul(d_Fc['XX'], d_XX) + np.matmul(d_Fc['YY'], d_YY)).flatten()) }, \
           { key : deriv for key,deriv in zip(deriv_keys, np.matmul(calc_d_loc_phase_func(gra, params_dic['dec'], f, d).T, d_gra_dec_psi(use_rot)[:2]).T) }

# location phase factor
def calc_d_loc_phase_func(gra, dec, f, d):
    theta = np.pi/2 - dec
    return loc_phase_func(gra, dec, f, d) * 1j * 2 * np.pi * f * np.array([
        # del_gra_Flp
        np.matmul(d, np.array([ -sin(gra)*sin(theta),  cos(gra)*sin(theta),           np.zeros_like(gra) ])),
        # del_dec_Flp
        np.matmul(d, np.array([ -cos(gra)*cos(theta), -sin(gra)*cos(theta), sin(theta)*np.ones_like(gra) ]))
        ])

# antenna patterns
def calc_d_ant_pat_funcs(D, XX, YY):
    # return d_Fp, d_Fc
    return {'XX': np.matmul(D,XX).T[None,:,:], 'YY':-np.matmul(D,YY).T[None,:,:]}, \
           {'XX': np.matmul(D,YY).T[None,:,:], 'YY': np.matmul(D,XX).T[None,:,:]}

# antenna pattern vectors
def calc_d_ant_pat_vectors(gra, dec, psi):
    # return del_gra_dec_psi_XX, del_gra_dec_psi_XX
    return np.array([ [ -cos(psi)*cos(gra) + sin(psi)*sin(gra)*sin(dec),  cos(psi)*sin(gra) + sin(psi)*cos(gra)*sin(dec),                   np.zeros_like(gra) ],
                      [                     -sin(psi)*cos(gra)*cos(dec),                      sin(psi)*sin(gra)*cos(dec), -sin(psi)*sin(dec)*np.ones_like(gra) ],
                      [  sin(psi)*sin(gra) - cos(psi)*cos(gra)*sin(dec),  sin(psi)*cos(gra) + cos(psi)*sin(gra)*sin(dec),  cos(psi)*cos(dec)*np.ones_like(gra) ] ]).T, \
           np.array([ [  sin(psi)*cos(gra) + cos(psi)*sin(gra)*sin(dec), -sin(psi)*sin(gra) + cos(psi)*cos(gra)*sin(dec),                   np.zeros_like(gra) ],
                      [                     -cos(psi)*cos(gra)*cos(dec),                      cos(psi)*sin(gra)*cos(dec), -cos(psi)*sin(dec)*np.ones_like(gra) ],
                      [  cos(psi)*sin(gra) + sin(psi)*cos(gra)*sin(dec),  cos(psi)*cos(gra) - sin(psi)*sin(gra)*sin(dec), -sin(psi)*cos(dec)*np.ones_like(gra) ] ]).T



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import numpy as np
import sympy as sp

import gwbench.basic_relations as brs
from gwbench.utils import time_fac, strain_fac

cos  = sp.cos
sin  = sp.sin
log  = sp.log
sqrt = sp.sqrt
PI   = np.pi

wf_symbs_string = 'f Mc eta chi1z chi2z DL tc phic iota lam_t delta_lam_t'
f, Mc, eta, chi1z, chi2z, DL, tc, phic, iota, lam_t, delta_lam_t = sp.symbols(wf_symbs_string, real=True)

#------from Anruadha--------
# defining constants
GammaE = 0.577215664901532

def hfpc(f, Mc, eta, chi1z, chi2z, DL, tc, phic, iota, lam_t, delta_lam_t, is_lam12=0):
    '''
    Mc ... in solar mass
    DL ... in mega parsec
    '''
    # convert to sec
    Mc = Mc * time_fac
    DL = DL * time_fac/strain_fac

    # get sym and asym chi combinations
    chi_s = brs.chi_s(chi1z,chi2z)
    chi_a = brs.chi_a(chi1z,chi2z)

    if is_lam12:
        # in this case: lam_t = lam1, delta_lam_t = lam2
        lam1 = lam_t
        lam2 = delta_lam_t
        delta = sqrt(1 - 4 * eta)
        lam_t = 8./13. * ( (1. + 7. * eta - 31. * eta**2) * (lam1 + lam2) +
                       delta * (1. + 9. * eta - 11. * eta**2) * (lam1 - lam2) )
        delta_lam_t = 0.5 * ( delta * (1319. - 13272. * eta + 8944. * eta**2) / 1319. * (lam1 + lam2)+
                       (1319. - 15910. * eta + 32850. * eta**2 + 3380. * eta**3) / 1319. * (lam1 - lam2) )

    '''
    Mc is in sec, e.g., Mc = 10*MTSUN_SI (for 10 solar mass)
    DL is in sec, e.g., DL = 100*1e6*PC_SI/C_SI (for 100 Mpc)
    '''
    M = Mc/eta**(3./5.)
    delta = (1.-4.*eta)**0.5
    v  = (PI*M*f)**(1./3.)
    flso = brs.f_isco(M)
    vlso = (PI*M*flso)**(1./3.)
    A =((5./24.)**0.5/PI**(2./3.))*(Mc**(5./6.)/DL)

    # 3.5PN phasing (point particle limit)
    p0 = 1.

    p1 = 0

    p2 = (3715./756. + (55.*eta)/9.)

    p3 = (-16.*PI + (113.*delta*chi_a)/3. + (113./3. - (76.*eta)/3.)*chi_s)

    p4 = (15293365./508032. + (27145.*eta)/504.+ (3085.*eta**2)/72. + (-405./8. + 200.*eta)*chi_a**2 - (405.*delta*chi_a*chi_s)/4. + (-405./8. + (5.*eta)/2.)*chi_s**2)

    gamma = (732985./2268. - 24260.*eta/81. - 340.*eta**2/9.)*chi_s + (732985./2268. + 140.*eta/9.)*delta*chi_a

    p5 = (38645.*PI/756. - 65.*PI*eta/9. - gamma)

    p5L = (38645.*PI/756. - 65.*PI*eta/9. - gamma)*3*log(v/vlso)

    p6 = (11583231236531./4694215680. - 640./3.*PI**2 - 6848./21.*GammaE + eta*(-15737765635./3048192. + 2255./12.*PI**2) + eta*eta*76055./1728. - eta*eta*eta*127825./1296. \
         - (6848./21.)*log(4.) + PI*(2270.*delta*chi_a/3. + (2270./3. - 520.*eta)*chi_s) + (75515./144. - 8225.*eta/18.)*delta*chi_a*chi_s \
         + (75515./288. - 263245.*eta/252. - 480.*eta**2)*chi_a**2 + (75515./288. - 232415.*eta/504. + 1255.*eta**2/9.)*chi_s**2)

    p6L = -(6848./21.)*log(v)

    p7 = (((77096675.*PI)/254016. + (378515.*PI*eta)/1512.- (74045.*PI*eta**2)/756. + (-25150083775./3048192. + (10566655595.*eta)/762048. - (1042165.*eta**2)/3024. + (5345.*eta**3)/36.
         + (14585./8. - 7270.*eta + 80.*eta**2)*chi_a**2)*chi_s + (14585./24. - (475.*eta)/6. + (100.*eta**2)/3.)*chi_s**3 + delta*((-25150083775./3048192.
         + (26804935.*eta)/6048. - (1985.*eta**2)/48.)*chi_a + (14585./24. - 2380.*eta)*chi_a**3 + (14585./8. - (215.*eta)/2.)*chi_a*chi_s**2)))

    phase = 2*f*PI*tc - phic - PI/4. + (3./(128.*v**5*eta))*(p0 + v*p1 + v**2*p2 + v**3*p3+ v**4*p4 + v**5*(p5+p5L) + v**6*(p6+p6L) + v**7*p7)

    # From Lesli Wade's paper ... adding the tidal deformability phase
    pT = (3./128./eta/v**5.)*(-39.*lam_t*v**10/2. + (-3115.*lam_t/64. + 6595.*(1-4*eta)**0.5*delta_lam_t/364.)*v**12)
    phase += pT

    hp = 0.5*(1+(cos(iota))**2)*A*f**(-7./6.)*(cos(phase) + 1j*sin(phase))
    hc = -1j*cos(iota)*A*f**(-7./6.)*(cos(phase) + 1j*sin(phase))

    return hp, hc



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


'''This module contains two methods that calculate numerical derivatives.
'''

import numdifftools as nd

import gwbench.analytic_derivatives as ad
import gwbench.utils as utils

def part_deriv_hf_func(hf, symbols_list, deriv_symbs_list, f, params_dic,
                       pl_cr=0, compl=1, label='hf',
                       ana_deriv_symbs_list=None, ana_deriv_aux=None,
                       step=1e-9, method='central', order=2, d_order_n=1):

    if 'f' in symbols_list: symbols_list.remove('f')
    if ana_deriv_symbs_list is None: _ana_deriv_symbs_list = []
    else:                            _ana_deriv_symbs_list = [ el for el in deriv_symbs_list if el in ana_deriv_symbs_list ]
    params_list       = [ params_dic[param] for param in symbols_list ]
    _deriv_symbs_list = [ el for el in deriv_symbs_list if el not in _ana_deriv_symbs_list ]
    deriv_params_list = [ params_dic[param] for param in _deriv_symbs_list ]

    if _ana_deriv_symbs_list:
        if pl_cr:
            d_hfpc         = { pc : ad.wf_ana_derivs(f, params_dic, ana_deriv_aux[f'hf{pc}']) for pc in ('p', 'c') }
            del_hf_dic_ana = { f'del_{name}_{label + pc}' : d_hfpc[pc][name] for name in _ana_deriv_symbs_list for pc in ('p', 'c') }
        else:
            d_hf           = ad.dr_ana_derivs(f, params_dic, **ana_deriv_aux)
            del_hf_dic_ana = { f'del_{name}_{label}'      : d_hf[name] for name in _ana_deriv_symbs_list }

    if not _deriv_symbs_list: return del_hf_dic_ana

    def hf_of_deriv_params(f, *deriv_params_list):
        return hf(f, *[deriv_params_list[_deriv_symbs_list.index(el)] if el in _deriv_symbs_list else params_list[i] for i,el in enumerate(symbols_list)])

    del_hf = part_deriv(hf_of_deriv_params, f, deriv_params_list, pl_cr, compl, step, method, order, d_order_n)

    if pl_cr:
        if len(_deriv_symbs_list) == 1: del_hf = (del_hf[0][:,None], del_hf[1][:,None])
        del_hf_dic = { f'del_{name}_{label + pc}' : _del_hf[:,i] for i,name in enumerate(_deriv_symbs_list) for pc, _del_hf in zip(('p', 'c'), del_hf) }
    else:
        if len(_deriv_symbs_list) == 1: del_hf = del_hf[:,None]
        del_hf_dic = { f'del_{name}_{label}'      :  del_hf[:,i] for i,name in enumerate(_deriv_symbs_list) }

    if _ana_deriv_symbs_list:
        del_hf_dic  = { **del_hf_dic, **del_hf_dic_ana }
        sorted_keys = [ key for name in deriv_symbs_list for key in del_hf_dic if name == key[4:].split(f'_{label}')[0] ]
        del_hf_dic  = { key : del_hf_dic[key] for key in sorted_keys}

    return del_hf_dic


def part_deriv(func, f, params_list, pl_cr=0, compl=None, step=1e-9, method='central', order=2, d_order_n=1):
    if pl_cr:
        if compl:
            def amp_pha_of_hfpc_compl(f, *params_list):
                return utils.pl_cr_to_amp_pha(*func(f, *params_list))

            amp_pl, pha_pl, amp_cr, pha_cr = amp_pha_of_hfpc_compl(f, *params_list)

            del_amp_pl = part_deriv_ndGradient(amp_pha_of_hfpc_compl, f, params_list, 0, step, method, order, d_order_n)
            del_pha_pl = part_deriv_ndGradient(amp_pha_of_hfpc_compl, f, params_list, 1, step, method, order, d_order_n)
            del_amp_cr = part_deriv_ndGradient(amp_pha_of_hfpc_compl, f, params_list, 2, step, method, order, d_order_n)
            del_pha_cr = part_deriv_ndGradient(amp_pha_of_hfpc_compl, f, params_list, 3, step, method, order, d_order_n)

            del_hfp = utils.z_deriv_from_amp_pha(amp_pl, pha_pl, del_amp_pl, del_pha_pl)
            del_hfc = utils.z_deriv_from_amp_pha(amp_cr, pha_cr, del_amp_cr, del_pha_cr)

            return del_hfp, del_hfc

        else:
            def amp_pha_of_hfpc_real(f, *params_list):
                return utils.amp_pha_from_re_im(*func(f, *params_list))

            amp, pha = amp_pha_of_hfpc_real(f, *params_list)
            del_amp = part_deriv_ndGradient(amp_pha_of_hfpc_real, f, params_list, 0, step, method, order, d_order_n)
            del_pha = part_deriv_ndGradient(amp_pha_of_hfpc_real, f, params_list, 1, step, method, order, d_order_n)

            return utils.re_im_from_z(utils.z_deriv_from_amp_pha(amp, pha, del_amp, del_pha))

    else:
        if compl:
            def amp_pha_of_hf(f, *params_list):
                return utils.amp_pha_from_z(func(f, *params_list))

            amp, pha = amp_pha_of_hf(f, *params_list)
            del_amp = part_deriv_ndGradient(amp_pha_of_hf, f, params_list, 0, step, method, order, d_order_n)
            del_pha = part_deriv_ndGradient(amp_pha_of_hf, f, params_list, 1, step, method, order, d_order_n)

            return utils.z_deriv_from_amp_pha(amp, pha, del_amp, del_pha)

        else:
            return part_deriv_ndGradient(func, f, params_list, None, step, method, order, d_order_n)


def part_deriv_ndGradient(func, f, params_list=None, funcid=None, step=1e-9, method='central', order=2, d_order_n=1):
    def wraps(x):
        if funcid == None: return func(f,*x)
        else:              return func(f,*x)[funcid]

    if params_list == None: return nd.Gradient(wraps, step=step, method=method, order=order, n=d_order_n)
    else:                   return nd.Gradient(wraps, step=step, method=method, order=order, n=d_order_n)(params_list)



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import os

import dill

import gwbench.antenna_pattern_np as ant_pat_np
import gwbench.antenna_pattern_sp as ant_pat_sp
import gwbench.waveform as wfc
import gwbench.wf_derivatives_num as wfd_num
import gwbench.wf_derivatives_sym as wfd_sym
import gwbench.utils as utils

lambdified_functions_path = os.path.join(os.getcwd(), 'lambdified_functions')

def calc_det_responses_derivs_num(loc, wf, deriv_symbs_string, f_arr,
                                  params_dic, use_rot=1, label='hf',
                                  step=1e-9, method='central', order=2, d_order_n=1,
                                  user_locs=None, ana_deriv_symbs_string=None,
                                  ana_deriv_aux=None):

    wf_symbs_list    = wf.wf_symbs_string.split(' ')
    deriv_symbs_list = deriv_symbs_string.split(' ')
    if ana_deriv_symbs_string is None: ana_deriv_symbs_list = None
    else:                              ana_deriv_symbs_list = ana_deriv_symbs_string.split(' ')

    if 'f' in wf_symbs_list: wf_symbs_list.remove('f')
    if 'f' in deriv_symbs_list: deriv_symbs_list.remove('f')

    if loc == None:
        wf_params_list = list(utils.get_sub_dict(params_dic, wf_symbs_list).values())

        def pc_func(f_arr, *wf_params_list):
            return wf.calc_wf_polarizations(f_arr, [ wf_params_list[wf_symbs_list.index(el)]
                                                     for el in wf_symbs_list ])

        return wfd_num.part_deriv_hf_func(pc_func, wf_symbs_list, deriv_symbs_list, f_arr, params_dic,
                                          pl_cr=1, compl=1, label=label,
                                          ana_deriv_symbs_list=ana_deriv_symbs_list,
                                          ana_deriv_aux=ana_deriv_aux,
                                          step=step, method=method, order=order, d_order_n=d_order_n)

    else:
        ap_symbs_list = ant_pat_np.ap_symbs_string.split(' ')
        if 'f' in ap_symbs_list: ap_symbs_list.remove('f')

        dr_symbs_list  = utils.reduce_symbols_strings(wf.wf_symbs_string,
                                                      ant_pat_np.ap_symbs_string).split(' ')
        dr_params_list = list(utils.get_sub_dict(params_dic, dr_symbs_list).values())

        def dr_func(f_arr, *dr_params_list):
            hfp, hfc    = wf.calc_wf_polarizations(f_arr, [ dr_params_list[dr_symbs_list.index(el)]
                                                            for el in wf_symbs_list ])
            Fp, Fc, Flp = ant_pat_np.antenna_pattern_and_loc_phase_fac(f_arr,
                *[ dr_params_list[dr_symbs_list.index(el)] for el in ap_symbs_list ],
                loc, use_rot, user_locs=user_locs)
            return Flp * (hfp * Fp + hfc * Fc)

        return wfd_num.part_deriv_hf_func(dr_func, dr_symbs_list, deriv_symbs_list, f_arr, params_dic,
                                          pl_cr=0, compl=1, label=label,
                                          ana_deriv_symbs_list=ana_deriv_symbs_list,
                                          ana_deriv_aux=ana_deriv_aux,
                                          step=step, method=method, order=order, d_order_n=d_order_n)



def generate_det_responses_derivs_sym(wf_model_name, wf_other_var_dic, deriv_symbs_string, use_rot,
                                      locs=None, user_locs=None, pl_cr=True, user_waveform=None,
                                      user_lambdified_functions_path=None, logger=None):

    # initialize waveform object
    wf = wfc.Waveform(wf_model_name, wf_other_var_dic, user_waveform=user_waveform)

    # check that derivative variables are a subset of the detector response variables
    full_set = set( wf.wf_symbs_string.split(' ') + ant_pat_np.ap_symbs_string.split(' ') )
    sub_set  = set( deriv_symbs_string.split(' ') )
    if not sub_set <= full_set:
        utils.log_msg('The choice of derivative variables is not a subset of the waveform ' +
                      'and antenna pattern variables!', logger=logger, level='ERROR')

    utils.log_msg( 'Generating lambdified derivatives via sympy with the following settings:',
                  logger=logger, level='DEBUG')
    utils.log_msg(f'    wf_model_name      = {wf.wf_model_name}', logger=logger, level='DEBUG')
    utils.log_msg(f'    deriv_symbs_string = {deriv_symbs_string}', logger=logger, level='DEBUG')
    utils.log_msg(f'    use_rot            = {bool(use_rot)}', logger=logger, level='DEBUG')
    utils.log_msg( '    Use these settings for a network loading these derivatives.',
                  logger=logger, level='DEBUG')

    # make sure specified locations are known
    if locs is None: locs = utils.available_locs
    else:
        for loc in locs:
            if loc not in utils.available_locs and loc not in user_locs:
                utils.log_msg(f'generate_det_responses_derivs_sym: Specified location {loc} not ' +
                               'known in antenna pattern module and was not provided in user_locs.',
                                logger=logger, level='ERROR')

    if user_lambdified_functions_path is None: output_path = lambdified_functions_path
    else:
        output_path = os.path.join(user_lambdified_functions_path, 'lambdified_functions')
    if not os.path.exists(output_path): os.makedirs(output_path)

    responses = { loc : None for loc in locs }
    hfpc      = wf.calc_wf_polarizations_expr()
    if pl_cr: responses['pl_cr'] = hfpc

    # compute sympy expressions of the detector responses
    # compute derivatives
    for key in responses.keys():
        if key == 'pl_cr':
            utils.log_msg(f'Generating lambdified derivatives for the plus/cross polarizations.',
                          logger=logger, level='INFO')
            utils.log_msg('    Calculating derivatives of the plus/cross polarizations.',
                          logger=logger, level='INFO')
            _deriv_symbs_string = utils.remove_symbols(deriv_symbs_string, wf.wf_symbs_string)
            symbs_string        = wf.wf_symbs_string
        else:
            utils.log_msg(f'Generating lambdified derivatives for {key}.',
                          logger=logger, level='INFO')
            utils.log_msg(f'    Loading the detector response expression for {key}.',
                          logger=logger, level='INFO')
            responses[key]      = ant_pat_sp.detector_response_expr(hfpc[0], hfpc[1], loc,
                                                                    use_rot, user_locs=user_locs)
            utils.log_msg(f'    Calculating derivatives of the detector responses for: {key}.',
                          logger=logger, level='INFO')
            _deriv_symbs_string = deriv_symbs_string
            symbs_string        = utils.reduce_symbols_strings(wf.wf_symbs_string,
                                                               ant_pat_np.ap_symbs_string)

        if not _deriv_symbs_string:
            utils.log_msg( '    The plus and cross polarizations of the waveform model ' +
                          f'{wf.wf_model_name}   do not depend on the derivative variables:  ' +
                          f'{deriv_symbs_string}\n' +
                          f'         Did not generate lambdified functions file!',
                          logger=logger, level='WARNING')
            continue

        deriv_dic                    = wfd_sym.part_deriv_hf_expr(responses[key], symbs_string,
                                                                  _deriv_symbs_string,
                                                                  pl_cr=(key=='pl_cr'))
        deriv_dic['variables']       = symbs_string
        deriv_dic['deriv_variables'] = _deriv_symbs_string

        file_name = os.path.join(
            output_path,
            f'par_deriv_WFM_{wf.wf_model_name}_' +
            f'DVAR_{_deriv_symbs_string.replace(" ", "_")}_ROT_{int(use_rot)}_DET_{key}.dat')
        utils.log_msg(f'    Stored at: {file_name}', logger=logger, level='INFO')
        with open(file_name, "wb") as fi:
            dill.dump(deriv_dic, fi, recurse=True)


def load_det_responses_derivs_sym(loc, wf_model_name, deriv_symbs_string, use_rot, gen_derivs=None,
                                  return_bin=0, user_lambdified_functions_path=None, logger=None):
    if user_lambdified_functions_path is None: _user_lambdified_functions_path = lambdified_functions_path
    file_name = f'par_deriv_WFM_{wf_model_name}_' + \
                f'DVAR_{deriv_symbs_string.replace(" ", "_")}_ROT_{int(use_rot)}_DET_{loc}.dat'

    try:
        with open(os.path.join(_user_lambdified_functions_path, file_name), "rb") as fi:
            if return_bin: return fi.read()
            else:          return dill.load(fi)
    except FileNotFoundError:
        if gen_derivs is None: utils.log_msg(f'Could not find the lambdified function file: {file_name}',
                                             logger=logger, level='ERROR')
        else:
            if loc == 'pl_cr': locs = []
            else:              locs = [loc]
            generate_det_responses_derivs_sym(wf_model_name, gen_derivs['wf_other_var_dic'],
                                              deriv_symbs_string, use_rot, locs=locs,
                                              user_locs=gen_derivs['user_locs'], pl_cr=gen_derivs['pl_cr'],
                                              user_waveform=gen_derivs['user_waveform'], logger=logger,
                                              user_lambdified_functions_path=user_lambdified_functions_path)
            return load_det_responses_derivs_sym(loc, wf_model_name, deriv_symbs_string, use_rot,
                                                 gen_derivs=None, return_bin=return_bin, logger=logger,
                                                 user_lambdified_functions_path=user_lambdified_functions_path)



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import numpy as np
from scipy.integrate import simps

#-----overlap function-------
def scalar_product_integrand(hf, gf, psd):
    temp = hf * np.conj(gf)
    return 2 * np.real((temp + np.conj(temp)) / psd)

def scalar_product_freq_array(hf, gf, psd, freqs):
    return simps(scalar_product_integrand(hf, gf, psd), freqs)

#-----SNR function-------
def snr_square_integrand(hf, psd):
    return 4 * np.divide(np.power(np.abs(hf), 2), psd)

def snr_square_freq_array(hf, psd, freqs):
    return simps(snr_square_integrand(hf, psd), freqs)

def snr_freq_array(hf, psd, freqs):
    return np.sqrt(snr_square_freq_array(hf, psd, freqs))

def snr_snr_sq_freq_array(hf, psd, freqs):
    snr_sq = simps(snr_square_integrand(hf, psd), freqs)
    return np.sqrt(snr_sq), snr_sq

#-----fft method from Anuradha-------
def rfft_normalized(time_series, dt, n=None):
    return np.fft.rfft(time_series, n) * dt

def fft_normalized(time_series, dt, n=None):
    return np.fft.fft(time_series, n) * dt
